/**
 * Copyright (c) 2012 - 2019 Data In Motion and others.
 * All rights reserved. 
 * 
 * This program and the accompanying materials are made available under the terms of the 
 * Eclipse Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     Data In Motion - initial API and implementation
 */
package org.gecko.core.api.connection;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ThreadFactory;

/**
 * A default connection facade, that allows execution of HTTP requests
 * without external dependencies.
 * @author Mark Hoffmann
 * @since 28.04.2020
 */
public abstract class ConnectionFacade {

	private final Map<Class<?>, DeSerializer<?>> serializerMap = new HashMap<>();
	private final Map<Class<?>, DeSerializer<?>> deserializerMap = new HashMap<>();
	private ThreadFactory threadFactory;
	private String connectionURL = null;
	private int timeout = 10000;

	public enum ConnectionMethod {
		GET,
		POST,
		PUT,
		DELETE
	}

	/**
	 * Constructor with default data and GUID parameter
	 * @param url the connection url
	 */
	public ConnectionFacade(String url) {
		setConnectionURL(url);
	}
	
	/**
	 * Sets the threadFactory.
	 * @param threadFactory the threadFactory to set
	 */
	public void setThreadFactory(ThreadFactory threadFactory) {
		this.threadFactory = threadFactory;
	}
	
	/**
	 * Returns the threadFactory.
	 * @return the threadFactory
	 */
	public ThreadFactory getThreadFactory() {
		return threadFactory;
	}

	/**
	 * Sets the deserializer.
	 * @param type the type to get the de-serializer for
	 * @param deserializer the deserializer to set
	 */
	public void setDeserializer(Class<?> type, DeSerializer<?> deserializer) {
		deserializerMap.put(type, deserializer);
	}
	
	/**
	 * Returns the deserializer.
	 * @param type the type to get the de-serializer for
	 * @return the deserializer. Can be <code>null</code>
	 */
	public DeSerializer<?> getDeserializer(Class<?> type) {
		if (type == null) {
			return null;
		}
		return deserializerMap.get(type);
	}
	
	/**
	 * Sets the serializer.
	 * @param type the type to get the serializer for
	 * @param serializer the serializer to set
	 */
	public void setSerializer(Class<?> type, DeSerializer<?> serializer) {
		serializerMap.put(type, serializer);
	}
	
	/**
	 * Returns the serializer.
	 * @param type the type to get the serializer for
	 * @return the serializer
	 */
	public DeSerializer<?> getSerializer(Class<?> type) {
		if (type == null) {
			return null;
		}
		return serializerMap.get(type);
	}

	/**
	 * Sends a GET request
	 * @param <T> the type of the return instance
	 * @param objectType the object type
	 * @param path the relative path
	 * @param properties some additional properties
	 * @return the instance of the response
	 */
	@SuppressWarnings("unchecked")
	public <T> T sendGetRequest(Class<T> objectType, String path, Map<String, Object> properties) {
		HttpURLConnection conn = null;
		try {
			// prepare GET request
			URL url = getConnectionURL();
			if (path != null) {
				url = new URL(getConnectionURL().toString() + path);
			}
			conn = prepareRequest(ConnectionMethod.GET, url);
			// load into EMF data model
			DataInputStream dis = new DataInputStream(conn.getInputStream());
			DeSerializer<?> deserializer = getDeserializer(objectType);
			if (deserializer != null && deserializer.canDeserialize()) {
				return (T)deserializer.deserialize(dis, null);
			}
		} catch (IOException e) {
			throw new IllegalStateException("I/O error executing GET request: " + path, e);
		} catch (Exception e) {
			throw new IllegalStateException("Error executing GET request: " + path, e);
		} finally {
			if (conn != null) {
				conn.disconnect();
			}
		}
		throw new IllegalStateException("Error executing GET request. This code should not be reached");
	}
	
	/**
	 * Sends a GET request
	 * @param <T> the type of the return instance
	 * @param objectType the object type
	 * @param path the relative path
	 * @param properties some additional properties
	 * @return the promise instance that resolves the response
	 */
//	public <T> Promise<T> sendGetRequestAsync(Class<T> objectType, String path, Map<String, Object> properties) {
//		PromiseFactory pf = createPromiseFactory();
//		return pf.submit(()->sendGetRequest(objectType, path, properties));
//	}

	/**
	 * Sends a get request and only returns its HTTP response code
	 * @return the HTTP response code
	 */
	public int sendGetRequestStatusOnly(){
		return sendRequest(ConnectionMethod.GET, null);
	}
	
	/**
	 * Sends a get request and only returns its HTTP response code 
	 * @return the HTTP promise, that resolves the response code
	 */
//	public Promise<Integer> sendGetRequestStatusAsync(){
//		PromiseFactory pf = createPromiseFactory();
//		return pf.submit(()->sendRequest(ConnectionMethod.GET, null));
//	}
	
	/**
	 * Sends a POST request
	 * @param object the object to be posted
	 * @return the HTTP response code or -1 on errors
	 */
	public int sendPostRequest(Object object, String path) {
		return sendPostRequest(object, path, null);
	}

	/**
	 * Sends a POST request
	 * @param object the object to be posted
	 * @return the HTTP response code or -1 on errors
	 */
	public int sendPostRequest(Object object, String path, Map<String, String> header) {
		return sendRequest(object, ConnectionMethod.POST, path, header, new HashMap<String, Object>());
	}

	/**
	 * Sends a PUT request
	 * @param object the object to be posted
	 * @return the HTTP response code or -1 on errors
	 */
	public int sendPutRequest(Object object, String path) {
		return sendRequest(object, ConnectionMethod.PUT, path, null, new HashMap<String, Object>());
	}

	/**
	 * Sends a DELETE request
	 * @return the HTTP response code or -1 on errors
	 */
	public int sendDeleteRequest() {
		return sendRequest(ConnectionMethod.DELETE, null);
	}

	/**
	 * Sends a request
	 * @param <T> The type of the object to upload
	 * @param object the instance
	 * @param method the HTTP method
	 * @param path the resource path
	 * @param header additional request headers
	 * @param properties additional properties
	 * @return the response code or throws an {@link IllegalStateException}
	 */
	@SuppressWarnings("unchecked")
	public <T> int sendRequest(T object, ConnectionMethod method, String path, Map<String, String> header, Map<String, Object> properties) {
		if (object == null) {
			throw new IllegalStateException("Error executing using method: " + method + " request because object or content type is/are null");
		}
		HttpURLConnection conn = null;
		try {
			URL url = getConnectionURL();
			if (path != null) {
				url = new URL(getConnectionURL().toString() + path);
			}
			// prepare POST request
			conn = prepareRequest(method, url, header);
			DeSerializer<T> serializer = (DeSerializer<T>) serializerMap.get(object != null ? object.getClass() : null);
			
			// add feedbacks to the resource
			// create output stream
			OutputStream os = conn.getOutputStream();
			DataOutputStream dos = new DataOutputStream(os);
			if (serializer != null && serializer.canSerialize()) {
				serializer.serialize((T)object, dos, properties);
			}
			int responseCode = conn.getResponseCode();
			if (responseCode == 200) {
				conn.getInputStream().close();
			}
			return responseCode;
		} catch (Exception e) {
			throw new IllegalStateException("Error executing '" + method + "' request: " + path, e);
		} finally {
			if (conn != null) {
				conn.disconnect();
			}
		}
	}


	/**
	 * Sends a request
	 * @param object the object to be whatever
	 * @param method the connection method, that should be executed
	 * @return the HTTP response code or -1 on errors
	 */
	@SuppressWarnings("unchecked")
	public <T> T  sendResultRequest(Class<T> resultType, Object object, ConnectionMethod method, String path, Map<String, String> header, Map<String, Object> properties) {
		if (object == null) {
			throw new IllegalStateException("Error executing '" + method + "' request because object or content type is/are null.");
		}
		HttpURLConnection conn = null;
		try {
			URL url = getConnectionURL();
			if (path != null) {
				path = URLEncoder.encode(path, "UTF-8");
				url = new URL(getConnectionURL().toString() + path);
			}
			// prepare POST request
			conn = prepareRequest(method, url, header);
			DeSerializer<T> serializer = (DeSerializer<T>) serializerMap.get(object != null ? object.getClass() : null);
			DeSerializer<T> deserializer = (DeSerializer<T>) deserializerMap.get(object != null ? object.getClass() : null);
			
			// add feedbacks to the resource
			// create output stream
			OutputStream os = conn.getOutputStream();
			DataOutputStream dos = new DataOutputStream(os);
			if (serializer != null && serializer.canSerialize()) {
				serializer.serialize((T)object, dos, properties);
			}
			int responseCode = conn.getResponseCode();
			if (responseCode == 200) {
				if (deserializer != null && deserializer.canDeserialize()) {
					return deserializer.deserialize(new DataInputStream(conn.getInputStream()), properties);
				}
				return null;
			}
			return null;
		} catch (Exception e) {
			throw new IllegalStateException("Error executing '" + method + "' request: " + path, e);
		} finally {
			if (conn != null) {
				conn.disconnect();
			}
		}
	}

	/**
	 * Sends a request without object
	 * @return the HTTP response code or -1 on errors
	 */
	public int sendRequest(ConnectionMethod method, URL url) {
		int responseCode = HttpURLConnection.HTTP_BAD_REQUEST;
		HttpURLConnection conn = null;
		try {
			// prepare request
			conn = prepareRequest(method, url);
			responseCode = conn.getResponseCode();      
			return responseCode;
		} catch (IOException e) {
			throw new IllegalStateException("Error executing '" + method + "' request: " + url, e);
		} finally {
			if (conn != null) {
				conn.disconnect();
			}
		}
	}

	/**
	 * Returns the timeout
	 * @return the timeout
	 */
	public int getTimeout() {
		return timeout;
	}

	/**
	 * Sets the timeout in milliseconds
	 * @param timeout the timeout in milliseconds
	 */
	public void setTimeout(int timeout) {
		this.timeout = timeout;
	}

	/**
	 * Sets the connection url
	 * @param url the connection url, must be not null
	 */
	public void setConnectionURL(String url) {
		if (url == null) {
			throw new IllegalArgumentException("Url parameter of the connection facade must be not null");
		}
		connectionURL = url;
	}

	/**
	 * Returns the connection uri
	 * @return the connection uri
	 * @throws MalformedURLException thrown id the uri string is wrong
	 */
	protected final URL getConnectionURL() throws MalformedURLException {
		return new URL(connectionURL);
	}

	/**
	 * Prepares a request for the given HTTP method
	 * @param method the HTTP method
	 * @return the {@link HttpURLConnection}
	 * @throws IOException
	 */
	protected HttpURLConnection prepareRequest(ConnectionMethod method, URL url) throws IOException {
		return prepareRequest(method, url, null);
	}

	/**
	 * Prepares a request for the given HTTP method
	 * @param method the HTTP method
	 * @return the {@link HttpURLConnection}
	 * @throws IOException
	 */
	protected HttpURLConnection prepareRequest(ConnectionMethod method, URL url, Map<String, String> header) throws IOException {
		url = url == null ? getConnectionURL() : url;
		HttpURLConnection conn = (HttpURLConnection) url.openConnection();
		Map<String, String> requestHeader = new HashMap<>(getHeaderMap());
		if (header != null) {
			requestHeader.putAll(header);
		}
		if (requestHeader != null) {
			requestHeader.forEach((k,v)->conn.setRequestProperty(k, v));
		}
		if (ConnectionMethod.PUT.equals(method) || ConnectionMethod.POST.equals(method)) {
			conn.setDoOutput(true);
		}
		conn.setRequestMethod(method.name());
		if (!ConnectionMethod.DELETE.equals(method)) {
			conn.setDoInput(true);
		}
		conn.setConnectTimeout(getTimeout());
		return conn;
	}

	/**
	 * Returns the header map
	 * @return the header map
	 */
	protected abstract Map<String, String> getHeaderMap();
	
	/**
	 * Creates a promise factory
	 * @return a promise factory
	 */
//	private PromiseFactory createPromiseFactory() {
//		ExecutorService es = threadFactory == null ? Executors.newSingleThreadExecutor() : Executors.newSingleThreadExecutor(threadFactory);
//		return new PromiseFactory(es);
//	}

}