/**
 * Copyright (c) 2012 - 2016 Data In Motion and others.
 * All rights reserved. 
 * 
 * This program and the accompanying materials are made available under the terms of the 
 * Eclipse Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     Data In Motion - initial API and implementation
 */
package org.gecko.core.tests;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.Collections;
import java.util.Set;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.osgi.framework.BundleContext;
import org.osgi.framework.Filter;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;
import org.osgi.util.promise.Promise;
import org.osgi.util.promise.PromiseFactory;
import org.osgi.util.tracker.ServiceTracker;

/**
 * Helper class that helps waiting for services to create, update or delete.
 * It adds support for blocking using {@link CountDownLatch}. Internally it uses the service tracker to
 * Track services
 * @author Mark Hoffmann
 * @param <T>
 * @since 10.11.2016
 */
public class ServiceChecker<T> {

	private final static Logger logger = Logger.getLogger(ServiceChecker.class.getName());

	// timeouts in seconds
	private final BundleContext context;
	private final PromiseFactory pf = new PromiseFactory(Executors.newFixedThreadPool(4), Executors.newScheduledThreadPool(4));
	private int createTimeout = 3000;
	private int modifyTimeout = 3000;
	private int removalTimeout = 3000;
	private ServiceTracker<T, T> tracker;
	private ServiceCheckerCustomizer<T, T> customizer = null;
	private Class<T> serviceClass = null;
	private Filter filter = null;
	private boolean running = false;
	private boolean immediate = false;
	private boolean verbose = false;
	// offset value, if there are already existing services, not to removed by our cleanup 
	private int removeOffset = 0;

	/**
	 * Set to <code>true</code>, to immediately close the checker, without waiting for removal.
	 * @param immediate the immediate to set
	 */
	public ServiceChecker<T> immediate(boolean immediate) {
		this.immediate = immediate;
		return this;
	}

	/**
	 * Set to an positive value, if there are existing services that
	 * are also tracked by this checker, and are not supposed to be removed.
	 * @param removeOffset the removeOffset to set
	 */
	public ServiceChecker<T> removeOffset(int removeOffset) {
		this.removeOffset = removeOffset > 0 ? removeOffset : 0;
		return this;
	}

	/**
	 * Returns the immediate.
	 * @return the immediate
	 */
	public boolean isImmediate() {
		return immediate;
	}

	/**
	 * Returns the verbose.
	 * @return the verbose
	 */
	public boolean isVerbose() {
		return verbose;
	}


	/**
	 * Returns the value for the timeout of a service creation
	 * @return the value for the timeout of a service creation
	 */
	public int getCreateTimeout() {
		return createTimeout;
	}

	/**
	 * Sets the timeout in seconds for a service creation
	 * @param createTimeout the value to be set
	 */
	public void setCreateTimeout(int createTimeout) {
		this.createTimeout = createTimeout;
	}

	/**
	 * Returns the value for the timeout of a service modification
	 * @return the value for the timeout of a service modification
	 */
	public int getModifyTimeout() {
		return modifyTimeout;
	}

	/**
	 * Sets the timeout in seconds for a service modification
	 * @param modifyTimeout the value to be set
	 */
	public void setModifyTimeout(int modifyTimeout) {
		this.modifyTimeout = modifyTimeout;
	}

	/**
	 * Returns the value for the timeout of a service removal
	 * @return the value for the timeout of a service removal
	 */
	public int getRemoveTimeout() {
		return removalTimeout;
	}

	/**
	 * Sets the timeout in seconds for a service removal
	 * @param removalTimeout the value to be set
	 */
	public void setRemovalTimeout(int removalTimeout) {
		this.removalTimeout = removalTimeout;
	}

	/**
	 * Creates a new instance.
	 * @param serviceClass the class to be tracked
	 * @param context the bundle context
	 */
	public ServiceChecker(Class<T> serviceClass, BundleContext context) {
		this.serviceClass = serviceClass;
		this.context = context;
	}

	/**
	 * Creates a new instance.
	 * @param filter the filter as string
	 * @param context the bundle context
	 * @throws InvalidSyntaxException
	 */
	public ServiceChecker(String filter, BundleContext context) throws InvalidSyntaxException {
		this.filter = context.createFilter(filter);
		this.context = context;
	}

	/**
	 * Creates a new instance.
	 * @param filter the filter 
	 * @param context the bundle context
	 */
	public ServiceChecker(Filter filter, BundleContext context) {
		this.filter = filter;
		this.context = context;
	}

	/**
	 * Starts the checker
	 */
	public void start() {
		if (context == null || (serviceClass == null && filter == null)) {
			fail("Error starting checker for service because service class or filter or bundle context is/are null");
		}
		if (running) {
			fail("Service check is already running");
		} else {
			running = true;
		}
		customizer = new ServiceCheckerCustomizer<T, T>(context, pf);
		customizer.setVerbose(verbose);
		if(serviceClass != null) {
			tracker = new ServiceTracker<T, T>(context, serviceClass, customizer);
		} else {
			tracker = new ServiceTracker<T, T>(context, filter, customizer);
		}
		tracker.open(true);
	}

	/**
	 * Starts the checker
	 */
	public ServiceChecker<T> run() {
		start();
		return this;
	}

	public ServiceChecker<T> verbose(boolean verbose) {
		this.verbose = verbose;
		if (customizer != null) {
			customizer.setVerbose(verbose);
		}
		return this;
	}

	public ServiceChecker<T> assertCreations(int expectedCreations, boolean wait) {
		if (!running) {
			start();
		}
		Promise<Boolean> p = customizer.resolveCreate(expectedCreations, wait ? getCreateTimeout() : 0);
		try {
			Boolean value = p.getValue().booleanValue();
			assertEquals(expectedCreations, getCurrentCreateCount());
			assertTrue(value.booleanValue());
		} catch (Exception e) {
			logger.log(Level.SEVERE, "Checking create for " + expectedCreations + " failed with exception", e);
			fail("Checking creations for " + expectedCreations + " failed with value " + getCurrentCreateCount() + " and exception " + e.getMessage());
		}
		return this;
	}

	public ServiceChecker<T> assertModifications(int expectedModifications, boolean wait) {
		if (!running) {
			start();
		}
		Promise<Boolean> p = customizer.resolveModify(expectedModifications, wait ? getModifyTimeout() : 0);
		try {
			Boolean value = p.getValue().booleanValue();
			assertEquals(expectedModifications, getCurrentModifyCount());
			assertTrue(value.booleanValue());
		} catch (Exception e) {
			logger.log(Level.SEVERE, "Checking modifications for " + expectedModifications + " failed with exception", e);
			fail("Checking modifications for " + expectedModifications + " failed with exception " + e.getMessage());
		}
		return this;
	}

	public ServiceChecker<T> assertRemovals(int expectedRemovals, boolean wait) {
		if (!running) {
			start();
		}
		Promise<Boolean> p = customizer.resolveRemove(expectedRemovals, wait ? getRemoveTimeout() : 0);
		try {
			Boolean value = p.getValue().booleanValue();
			assertEquals(expectedRemovals, getCurrentRemoveCount());
			assertTrue(value.booleanValue());
		} catch (Exception e) {
			logger.log(Level.SEVERE, "Checking removals for " + expectedRemovals + " failed with exception", e);
			fail("Checking removals for " + expectedRemovals + " failed with exception " + e.getMessage());
		}
		return this;
	}

	/**
	 * Stops the service checker
	 */
	public void stop() {
		if (tracker != null) {
			tracker.close();
			tracker = null;
		}
		if (customizer != null) {
			customizer.dispose();
			customizer = null;
		}
		if (running) {
			running = false;
		}
	}

	/**
	 * Returns the tracked service, that maybe <code>null</code>. It waits max 1s for the service to come
	 * @return the tracked service, that maybe <code>null</code>
	 */
	public T getTrackedService() {
		failIfNotRunning("Getting service");
		try {
			return tracker.waitForService(1000l);
		} catch (InterruptedException e) {
			throw new IllegalStateException(e);
		}
	}

	public ServiceChecker<T> trackedServiceNotNull() {
		if (!running) {
			start();
		}
		assertNotNull(getTrackedService());
		return this;
	}

	public ServiceChecker<T> trackedServiceNull() {
		if (!running) {
			start();
		}
		assertNull(getTrackedService());
		return this;
	}

	/**
	 * Returns the highest ranked tracked service reference, that maybe <code>null</code>
	 * @return the tracked service reference, that maybe <code>null</code>
	 */
	public ServiceReference<T> getTrackedServiceReference() {
		failIfNotRunning("Getting service reference");
		try {
			ServiceReference<T> ref = null;
			long timeout = 1000;
			long timeToWait = 5;
			do {
				Thread.sleep(timeToWait);
				timeout -= timeToWait;
				ref = tracker.getServiceReference();
				if (timeout <= 0) {
					break;
				}
			} while (ref == null);
			return ref;
		} catch (InterruptedException e) {
			throw new IllegalStateException(e);
		}
	}

	public Set<ServiceReference<T>> getAllServiceReferences() {
		failIfNotRunning("Getting all service references");
		if (customizer != null) {
			return customizer.getReferences();
		}
		return Collections.emptySet();
	}

	/**
	 * Returns the current count, how many times the create of a service was triggered
	 * @return the number of service creations
	 */
	public int getCurrentCreateCount() {
		failIfNotRunning("Current creation count");
		if (customizer != null) {
			return customizer.getAddCount();
		} else {
			throw new IllegalStateException("No customizer was created. This must be an error.");
		}
	}

	/**
	 * Returns the current count, how many times the modification of a service was triggered
	 * @return the number of service modifications
	 */
	public int getCurrentModifyCount() {
		failIfNotRunning("Current modification count");
		if (customizer != null) {
			return customizer.getModifyCount();
		} else {
			throw new IllegalStateException("No customizer was created. This must be an error.");
		}
	}

	/**
	 * Returns the current count, how many times the removal of a service was triggered
	 * @return the number of service removals
	 */
	public int getCurrentRemoveCount() {
		failIfNotRunning("Current removal count");
		if (customizer != null) {
			return customizer.getRemoveCount();
		} else {
			throw new IllegalStateException("No customizer was created. This must be an error.");
		}
	}

	public void awaitRemovals() {
		if (running) {
			Promise<Boolean> shutdown = customizer.resolveShutdown(removeOffset);
			try {
				shutdown.timeout(getRemoveTimeout()).getValue();
			} catch (Exception e) {
				logger.log(Level.SEVERE, "Error resolving the shutdown with exception", e);
			}
		}
	}

	/**
	 * This method checks for state running and provides the given message
	 * @param message the fail message
	 */
	private void failIfNotRunning(String message) {
		if (!running) {
			fail("Checker is not running. Did you call start? " + message);
		}
	}

}
